# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License
import os

SERVICE_TYPE = "_sdk._tcp.local."
DISCOVER_SERVICE_TYPE = "_spyder_remote_discover._tcp.local."
KERNEL_SERVICE_TYPE = "_spyder_remote_kernel._tcp.local."

# Number of cores
AUTO = 0

HERE = os.path.abspath(os.path.dirname(__file__))
PID_FILE = "daemon.pid"
PID_FILE_PATH = os.path.join(HERE, PID_FILE)



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License
"""
Conda API.
"""

import json
import os
import subprocess
import sys

from spyder_remote_server.utils import get_conda_info


class CondaManager:
    """
    A manager providing an API to the conda package and environment manager.
    """

    def __init__(self):
        # Conda config values
        self.CONDA_PREFIX = None
        self.ROOT_PREFIX = None
        self._envs_dirs = None
        self._pkgs_dirs = None
        self._user_agent = None
        self._proxy_servers = None
        self._conda_version = None

        # Process
        self._process = None
        self._stdout = None
        self._stderr = None

        self.set_conda_prefix(info=get_conda_info())
        self.user_rc_path = os.path.abspath(os.path.expanduser("~/.condarc"))
        self.sys_rc_path = os.path.join(self.ROOT_PREFIX, ".condarc")

    def _call_conda(
        self, extra_args, abspath=True, parse=False, callback=None, environ=None
    ):
        """
        Call conda with the list of extra arguments, and return the worker.

        The result can be force by calling worker.communicate(), which returns
        the tuple (stdout, stderr).

        Parameters
        ----------
        extra_args: list
            List of additional arguments to pass to conda.
        abspath: bool, optional
            Default is ``True``.
        parse: bool, optional
            Parse result as JSON. Default is ``False``.
        callback: callable or None, optional
            Any callback to use after processing results. Default is ``None``.
        environ: dict or None, optional
            The process environment dictionary. Default is ``None``.

        Returns
        -------
        subprocess.Popen
            Conda process object.
        """
        if abspath:
            if sys.platform == "win32":
                python = os.path.join(self.ROOT_PREFIX, "python.exe")
                conda = os.path.join(self.ROOT_PREFIX, "Scripts", "conda-script.py")
            else:
                python = os.path.join(self.ROOT_PREFIX, "bin/python")
                conda = os.path.join(self.ROOT_PREFIX, "bin/conda")
            cmd_list = [python, conda]
        else:
            # Just use whatever conda is on the path
            cmd_list = ["conda"]

        cmd_list.extend(extra_args)
        process = subprocess.Popen(
            cmd_list, env=environ, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        return process

    def _call_and_parse(self, extra_args, abspath=True, callback=None, environ=None):
        """
        Call conda process and parse results as JSON.

        Parameters
        ----------
        extra_args: list
            List of additional arguments to pass to conda.
        abspath: bool, optional
            Default is ``True``.
        callback: callable or None, optional
            Any callback to use after processing results. Default is ``None``.
        environ: dict or None, optional
            The process environment dictionary. Default is ``None``.

        Returns
        -------
        dict
            Dictionary with result information from conda process.
        """
        process = self._call_conda(
            extra_args,
            abspath=abspath,
            parse=True,
            callback=callback,
            environ=environ,
        )
        stdout, stderr = process.communicate()
        try:
            data = json.loads(stdout)
        except Exception:
            data = None

        return data

    def _set_environment_variables(self, prefix=None, no_default_python=False):
        """
        Set the right CONDA_PREFIX environment variable.

        Parameters
        ----------
        prefix: str or None, optional
            Full path to envirtonment. Default is ``None``.
        no_default_python: bool, optional
            Default is ``False``.

        Returns
        -------
        dict
            Process environment dictionary for given ``prefix``.
        """
        environ_copy = os.environ.copy()
        conda_prefix = self.ROOT_PREFIX
        if prefix:
            conda_prefix = prefix

        if conda_prefix:
            if conda_prefix == self.ROOT_PREFIX:
                name = "root"
            else:
                name = os.path.basename(conda_prefix)
            environ_copy["CONDA_PREFIX"] = conda_prefix
            environ_copy["CONDA_DEFAULT_ENV"] = name

        if no_default_python:
            environ_copy["CONDA_DEFAULT_PYTHON"] = None

        return environ_copy

    def set_conda_prefix(self, info=None):
        """
        Set the prefix of the conda environment.

        This function should only be called once (right after importing
        conda_api).

        Parameters
        ----------
        info: dict or None, optional
            Dictionary with conda informationa. Default is ``None``.
        """
        if info is None:
            # Find some conda instance, and then use info to get 'root_prefix'
            worker = self.info(abspath=False)
            info = worker.communicate()[0]
        else:
            self.ROOT_PREFIX = info["root_prefix"]
            self.CONDA_PREFIX = info["conda_prefix"]
            self._envs_dirs = info["envs_dirs"]
            self._pkgs_dirs = info["pkgs_dirs"]
            self._user_agent = info["user_agent"]

            version = []
            for part in info["conda_version"].split("."):
                try:
                    new_part = int(part)
                except ValueError:
                    new_part = part

                version.append(new_part)

            self._conda_version = tuple(version)

    def environment_exists(self, name=None, prefix=None, abspath=True, log=True):
        """
        Check if an environment exists by 'name' or by 'prefix'.

        If query is by 'name' only the default conda environments directory is
        searched.

        Parameters
        ----------
        name: str or None, optional
            Name of envirtonment. Default is ``None``.
        prefix: str or None, optional
            Full path to envirtonment. Default is ``None``.
        abspath: bool, optional
            Default is ``True``.
        log: bool, optional
            Default is ``True``.

        Returns
        -------
        bool
            Wheter the conda environment exists.
        """
        if name and prefix or (name is None and prefix is None):
            raise TypeError("Exactly one of 'name' or 'prefix' is required.")

        if name:
            prefix = self.get_prefix_envname(name)

        if prefix is None:
            prefix = self.ROOT_PREFIX

        return os.path.isdir(os.path.join(prefix, "conda-meta"))

    def info(
        self,
        prefix=None,
        abspath=True,
        unsafe_channels=False,
        all_=False,
    ):
        """
        Return a dictionary with configuration information.

        Parameters
        ----------
        prefix: str or None, optional
            Full path to envirtonment prefix from which to list packages.
            Default is ``None``.
        abspath: bool, optional
            Default is ``True``.
        unsafe_channels: bool, optionas
            Default is ``False``.
        all_: str, optional
            Default is ``False``.

        Returns
        -------
        dict
            Dictionary of conda information.
        """
        environ = self._set_environment_variables(prefix)
        cmd_list = ["info", "--json"]

        if unsafe_channels:
            cmd_list.extend(["--unsafe-channels"])

        if all_:
            cmd_list.extend(["--all"])

        return self._call_and_parse(
            cmd_list,
            abspath=abspath,
            environ=environ,
        )

    def list_packages(
        self,
        prefix=None,
        abspath=True,
        unsafe_channels=False,
    ):
        """
        Return a dictionary with configuration information.

        Parameters
        ----------
        prefix: str or None, optional
            Full path to envirtonment prefix from which to list packages.
            Default is ``None``.
        abspath: bool, optional
            Default is ``True``.
        unsafe_channels: bool, optionas
            Default is ``False``.

        Returns
        -------
        dict
            Dictionary of installed packages.
        """
        environ = self._set_environment_variables(prefix)
        cmd_list = ["list", "-p", prefix, "--json"]
        return self._call_and_parse(
            cmd_list,
            abspath=abspath,
            environ=environ,
        )

    def get_envs(self):
        """
        Return environment list of absolute path to their prefixes.

        Returns
        -------
        list
            list of absolute path to their prefixes. This does not include the
            `base` environment.
        """
        all_envs = []
        for env in self.envs_dirs:
            if os.path.isdir(env):
                envs_names = os.listdir(env)
                all_envs += [os.sep.join([env, i]) for i in envs_names]

        valid_envs = [
            env
            for env in all_envs
            if os.path.isdir(env) and self.environment_exists(prefix=env)
        ]

        return valid_envs

    @property
    def envs_dirs(self):
        """
        Conda environment directories.

        The first writable item should be used.

        Returns
        -------
        list
            List of path where environments are created.
        """
        if self._envs_dirs:
            envs_dirs = self._envs_dirs
        else:
            # Legacy behavior
            envs_path = os.sep.join([self.ROOT_PREFIX, "envs"])
            user_envs_path = os.sep.join([get_home_dir(), ".conda", "envs"])
            envs_dirs = [envs_path, user_envs_path]

        return envs_dirs


if __name__ == "__main__":
    cm = CondaManager()
    print(cm.get_envs())



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License
# https://github.com/torfsen/python-systemd-tutorial
# https://daemoniker.readthedocs.io/en/latest/
import os
import tempfile
import time

from daemoniker import Daemonizer
from daemoniker import SignalHandler1
from spyder_remote_server.constants import PID_FILE_PATH
from spyder_remote_server.server import SpyderRemoteServer


def do_things_here():
    print("Before daemonization")


def parent_only_code():
    print("Parent yei!")


def code_continues_here():
    print("code continues here!")
    server = SpyderRemoteServer()
    try:
        server.start_server()
    except Exception as e:
        server.stop()
    finally:
        server.stop()


with Daemonizer() as (is_setup, daemonizer):
    if is_setup:
        # This code is run before daemonization.
        # TODO: This has to run as `root`
        do_things_here()

    # We need to explicitly pass resources to the daemon; other variables
    # may not be correct
    is_parent = daemonizer(PID_FILE_PATH)

    if is_parent:
        # Run code in the parent after daemonization
        parent_only_code()


# We are now daemonized, and the parent just exited.


# Create a signal handler that uses the daemoniker default handlers for
# ``SIGINT``, ``SIGTERM``, and ``SIGABRT``
sighandler = SignalHandler1(PID_FILE_PATH)
sighandler.start()

# Or, define your own handlers, even after starting signal handling
def handle_sigint(signum):
    print("SIGINT received.")


sighandler.sigint = handle_sigint

print("code continues here!")
server = SpyderRemoteServer()
try:
    server.start_server()
except Exception as e:
    server.stop()
finally:
    server.stop()



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License
import os

# Constants
HERE = os.path.abspath(os.path.dirname(__file__))
CONFIG_TEMPLATE = "config.jinja2"
CONFIG_TEMPLATE_PATH = os.path.join(HERE, "config.jinja2")



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License
import multiprocessing

from spyder_remote_server.config import create_config
from spyder_remote_server.config import remove_config


def register_service():
    """
    TODO:
    """


def get_cpu_count(cores=None):
    """
    TODO:
    """
    cpu_count = multiprocessing.cpu_count()
    cores = cores or cpu_count

    if cores <= cpu_count:
        cpu_count = cores
    elif cores == 0:
        cpu_count = cpu_count - 1
    else:
        # cores > cpu_count, using system cpu_count
        pass

    return cpu_count


def install_daemon(guest, cores):
    """
    TODO:
    """
    create_config(guest, cores)
    register_service()


def uninstall_daemon():
    """
    TODO:
    """
    remove_config()



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License
"""
Spyder remote server utilities.
"""

import json
import os
import pwd
import socket
import subprocess
import sys
from contextlib import closing

# Constants
WIN = os.name == "nt"


def is_base_conda():
    """Check if current env is base enviroment."""
    return True


def is_root():
    """Check that current process is running as `root`."""
    return os.geteuid() == 0


def demote(user_uid, user_gid):
    """
    Callback to use with subprocess `preexec_fn` to change the user running
    the process.

    Parameters
    ----------
    user_uid: int
        TODO:
    user_gid: int
        TODO:
    """

    def result():
        report_ids("starting demotion")
        os.setgid(user_gid)
        os.setuid(user_uid)
        report_ids("finished demotion")

    return result


def report_ids(msg):
    print("uid, gid = %d, %d; %s" % (os.getuid(), os.getgid(), msg))


def run_process(cmd_list):
    """
    Run subprocess with cmd_list and return stdour, stderr, error.

    Returns
    -------
    tuple
        Tuple of stdout, stderr and any exception found.
    """
    stdout = ""
    stderr = ""
    error = False
    try:
        p = subprocess.Popen(cmd_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        stdout = stdout.decode()
        stderr = stderr.decode()
    except OSError:
        error = True

    return stdout, stderr, error


def is_conda_available():
    """
    Check if conda is available in path.

    Returns
    -------
    bool
        Wheter conda is available in the system.
    """
    return bool(get_conda_cmd_path())


def get_conda_cmd_path():
    """
    Check if conda is found on path.

    Returns
    -------
    str
        Path to conda executable.
    """
    cmds = []
    conda_path = None
    bin_folder = "Scripts" if WIN else "bin"
    conda_exe = "conda-script.py" if WIN else "conda"
    env_prefix = os.path.dirname(os.path.dirname(sys.prefix))

    cmds.append(os.path.join(env_prefix, bin_folder, conda_exe))
    cmds.append(os.path.join(sys.prefix, bin_folder, conda_exe))
    cmds.append("conda")

    for cmd in cmds:
        cmd_list = [cmd, "--version"]
        stdout, stderr, error = run_process(cmd_list)
        if not error:
            if stdout.startswith("conda ") or stderr.startswith("conda "):
                conda_path = cmd
                break

    return conda_path


def get_conda_info():
    """
    Return conda info as a dictionary.

    Returns
    -------
    dict
        Conda information.
    """
    conda_cmd = get_conda_cmd_path()
    info = None
    if conda_cmd:
        cmd_list = [conda_cmd, "info", "--json"]
        out, err, error = run_process(cmd_list)
        try:
            info = json.loads(out)
        except Exception:
            pass

    return info


def get_username():
    """
    Get the current username.

    https://stackoverflow.com/a/2899055
    """
    return pwd.getpwuid(os.getuid())[0]


def get_ip():
    """
    Get the public IP interface of the current machine.

    Returns
    -------
    str
        IP of current machine.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()

    return IP


def find_free_port():
    """
    Find free port.

    Returns
    -------
    int
        Free port found.
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License

"""
Spyder remote server setup.
"""

import ast
import os

from setuptools import find_packages
from setuptools import setup

# Constants
HERE = os.path.abspath(os.path.dirname(__file__))


def get_version(module="spyder_remote_server"):
    """Get version."""
    with open(os.path.join(HERE, module, "__init__.py"), "r") as f:
        data = f.read()

    lines = data.split("\n")
    for line in lines:
        if line.startswith("__version__"):
            version = ast.literal_eval(line.split("=")[-1].strip())
            break

    return version


def get_description():
    """Get long description."""
    with open(os.path.join(HERE, "README.md"), "r") as f:
        data = f.read()
    return data


setup(
    name="spyder-remote-server",
    version=get_version(),
    url="https://github.com/Semi-ATE/spyder-remote/tree/master/spyder-remote-server",
    description="Spyder remote server to connect to Spyder kernels via zeroconf.",
    long_description=get_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    install_requires=[
        "click",
        "daemoniker",
        "jinja2",
        # "psutil",
        "pyzmq",
        # "netifaces",
        "zeroconf <=0.27",
    ],
    entry_points={
        "console_scripts": [
            "spyder-remote-server = spyder_remote_server.cli:main",
        ]
    },
    include_package_data=True,
)



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License

"""
Spyder remote server in charge of launching the zeroconf service and provide
a socket via ZMQ to provide spyder-kernels start/stop.
"""

import fnmatch
import json
import logging
import os
import platform
import pwd
import signal
import socket
import subprocess
import sys
import tempfile
import time

import zmq
from zeroconf import IPVersion, ServiceInfo, Zeroconf

from spyder_remote_server.conda_api import CondaManager
from spyder_remote_server.config import get_log_path, read_config
from spyder_remote_server.constants import SERVICE_TYPE
from spyder_remote_server.utils import find_free_port, get_ip, demote


# TODO: Check logging
logging.basicConfig(
    filename=get_log_path(),
    filemode="a",
    format="%(asctime)s: %(name)s - %(levelname)s - %(message)s",
    level=logging.DEBUG,
)
logger = logging.getLogger(__name__)


class SpyderRemoteServer:
    """
    Spyder remote server in charge of launching the zeroconf service and provide
    a socket via ZMQ to provide spyder-kernels start/stop.
    """

    def __init__(self):
        # Connect signals to stop handling method
        signal.signal(signal.SIGINT, self.stop_server)
        signal.signal(signal.SIGTERM, self.stop_server)

        # Spyder kernels
        self._busy = None
        self._running = True
        self._kernels = {}
        self._current_properties = {}

        # Conda
        self._conda_manager = CondaManager()

        # Daemon server
        self._config = read_config()
        self._server_port = None

        # Zeroconf
        self._zeroconf = Zeroconf()
        self._service_port = find_free_port()
        self._service_info = None

    def get_properties(self):
        """
        Return properties to include in the zeroconf service. These properties
        include conda environments and local configuration values.

        Returns
        -------
        dic
            Properties dictionary.
        """
        envs = self._conda_manager.get_envs()
        valid_envs = set()
        valid_envs_has_kernels = set()
        for env_path in envs:
            env_name = os.path.basename(env_path)
            checks = []
            for pattern in self._config["exclude_environments"]:
                checks.append(not fnmatch.fnmatch(env_name, pattern))

            if all(checks):
                print(env_path)
                packages = self._conda_manager.list_packages(env_path)
                for item in packages:
                    check = item["dist_name"].startswith("spyder-kernels")
                    if check:
                        valid_envs_has_kernels.add(env_path)
                        break

                valid_envs.add(env_path)

        properties = {
            "server_port": self._server_port,
        }
        for i, env in enumerate(sorted(valid_envs)):
            if env in valid_envs_has_kernels:
                suffix = "yes"
            else:
                suffix = "no"

            properties[f"conda_env_{i}_{suffix}"] = env

        properties["guest_account"] = self._config["guest_account"]
        properties["name"] = self._config["name"]
        properties["max_kernels"] = self._config["max_kernels"]
        properties["guest_can_manage_environments"] = self._config[
            "guest_can_manage_environments"
        ]
        kernel_count = 0
        for env, kernel_data in self._kernels.items():
            kernel_count += len(kernel_data)

        properties["current_kernels"] = kernel_count

        return properties

    def get_service_info(self, properties=None):
        """
        Return the zeroconf service information.

        Returns
        -------
        zeroconf.ServiceInfo
            A zeroconf service object.
        """
        name = socket.gethostname() or platform.node()
        full_name = f"{name}.{SERVICE_TYPE}"
        if properties is None:
            self._current_properties = self.get_properties()
            properties = self._current_properties

        self._service_info = ServiceInfo(
            SERVICE_TYPE,
            full_name,
            port=self._service_port,
            properties=properties,  # key / value 255 len
            addresses=[socket.inet_aton(get_ip())],
        )
        return self._service_info

    def register_service(self):
        """
        Register zeroconf spyder remote service.
        """
        self._zeroconf.register_service(self.get_service_info())
        logger.info("Service registered!")
        print("Service registered!")

    def update_service(self):
        """
        Update service with updated information.
        """
        kernel_count = 0
        for _env, kernel_data in self._kernels.items():
            kernel_count += len(kernel_data)

        self._current_properties["current_kernels"] = kernel_count
        # self._zeroconf.update_service(self.get_service_info(self._current_properties))

    def start_kernel(self, prefix):
        """
        Start spyder-kernel located at environment prefix.

        Parameters
        ----------
        prefix: str
            Full path to conda environment prefix.
        """
        # FIXME: Use conda activation scripts
        python_path = f"{prefix}/bin/python"
        _, json_path = tempfile.mkstemp(suffix=".json")
        os.remove(json_path)
        cmd_args = [
            python_path,
            "-m",
            "spyder_kernels.console",
            f"--ip={get_ip()}",
            f"-f={json_path}",
        ]

        # To run as the guest user
        user_name = self._config["guest_account"]

        try:
            pw_record = pwd.getpwnam(user_name)
        except KeyError:
            print(f"User `{user_name}` not found!")
            return
        except Exception as err:
            print(f"Exception:\n {err}")
            return

        user_name = pw_record.pw_name
        user_home_dir = pw_record.pw_dir
        user_uid = pw_record.pw_uid
        user_gid = pw_record.pw_gid
        cwd = user_home_dir
        env = os.environ.copy()
        env["HOME"] = user_home_dir
        env["LOGNAME"] = user_name
        env["PWD"] = cwd
        env["USER"] = user_name
        try:
            kernel_proc = subprocess.Popen(
                cmd_args,
                preexec_fn=demote(user_uid, user_gid),
                cwd=cwd,
                env=env,
            )
        except Exception as error:
            print(error)

        if prefix not in self._kernels:
            self._kernels[prefix] = []

        time.sleep(1)
        while True:
            if os.path.isfile(json_path):
                with open(json_path, "r") as fh:
                    data = json.loads(fh.read())

                break
            time.sleep(1)

        self._kernels[prefix].append(
            {"process": kernel_proc, "kernel_data": data, "json_path": json_path}
        )
        print(f"The kernel process id: {kernel_proc.pid}")
        self.update_service()
        return data

    def stop_kernel(self, env=None):
        """
        Stop all kernels for given env if provided or all started kernels in no
        env is provided.
        """

    def start_server(self):
        """
        Start Spyder Remote server.
        """
        if not self._config["enable"]:
            print("Daemon not enabled. Check configuration!")
            return

        # self.start_kernel('/Users/goanpeca/miniconda3/envs/zeroconf')
        logger.info("Starting daemon!")
        logging.info("Starting daemon!")
        context = zmq.Context()
        socket = context.socket(zmq.REP)

        self._server_port = find_free_port()
        socket.bind(f"tcp://*:{self._server_port}")
        print(f"Bound to port {self._server_port}")
        self.register_service()

        while self._running:
            self._busy = True
            message = socket.recv()

            try:
                reply = self.process_request(message)
            except Exception as err:
                reply = json.dumps({"error": str(err)})
                print(err)

            socket.send_string(reply)
            self._busy = False

    def process_request(self, message_byte_string):
        """
        Parameters
        ----------
        message_byte_string: bytes
            The request to process as a bytes string. When decoded it should
            provide a parseable JSON string.

        Examples
        --------
        >>> message_example_data = {
            "kernel": {
                "command": "start",
                "prefix": "conda-prefix",
            },
            "conda": {
                "command": "create",
                "prefix": "conda-env prefix",
                "name": "conda-env name",
                "packages": [],
            },
        }

        Returns
        -------
        str
            A JSON dumped string.
        """
        message = json.loads(message_byte_string.decode("utf-8"))
        print("Received request: %s" % message)

        reply = {"test": "testing"}
        if "kernel" in message:
            reply = self.handle_kernel_request(message)
        elif "conda" in message:
            reply = self.handle_conda_request(message)

        return json.dumps(reply)

    def handle_kernel_request(self, message):
        """
        Parameters
        ----------
        message: dict
            The message with kernel commands. {

        Example
        -------
        >>> {
            "kernel": {
                "command": "start",  # stop, close_all
                "prefix": "<prefix-to-conda-env>",
            }
        }

        Returns
        -------
        dict
            Response with a "json_data" key.
        """
        kernel_message = message["kernel"]
        command = kernel_message["command"]
        prefix = kernel_message.get("prefix", "")

        json_data = {}
        if command == "start" and prefix and os.path.isdir(prefix):
            json_data = self.start_kernel(prefix)
        elif command == "close_all":
            json_data = self.kill_running_kernels()

        return {"json_data": json_data}

    def handle_conda_request(self, message):
        """
        Handle conda requests to the server.

        Parameters
        ----------
        message: dict
            The message with kernel commands.

        Returns
        -------
        dict
            Response.
        """
        return {"status": "ok"}

    def kill_running_kernels(self):
        """
        Kill any kernels started by this server.

        Returns
        -------
        dict
            Response with a "status" key.
        """
        for env, kernels_data in self._kernels.items():
            for kernel_data in kernels_data:
                json_path = kernel_data["json_path"]
                os.remove(json_path)

                proc = kernel_data["process"]
                try:
                    print(f"Killing env: {env}")
                    logger.info(f"Killing env: {env}")
                    proc.terminate()
                    _stdout, _stderr = proc.communicate()
                except Exception as e:
                    logger.error(str(e))

        self._kernels = {}

        return {"status": "ok"}

    def stop_server(self, *args):
        """
        Handle sigterm signal to unregister services and kill any started
        kernels.
        """
        logger.info("Stopping Spyder Remote daemon!")
        self._running = False
        try:
            self._zeroconf.unregister_service(self._service_info)
        except Exception as e:
            logger.error(str(e))

        self.kill_running_kernels()
        sys.exit(0)


if __name__ == "__main__":
    server = SpyderRemoteServer()
    server.start_server()



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License
"""
Load and parse spyder-remote configuration.
"""

import ast
import configparser
import multiprocessing
import os
import platform
import socket
import sys

from jinja2 import Environment, PackageLoader, select_autoescape
from spyder_remote_server.constants import AUTO
from spyder_remote_server.templates import CONFIG_TEMPLATE

# Constants
INIFILE = "spyder-remote-server.ini"
DEFAULT_CONFIG = {
    ##############################################################################
    # The `enable` keyword will determine if the Spyder Kernels Daemon will start
    #
    # Notes:
    #  * if `enable` is not supplied, it is presumed turned off!
    #  * if `enable` can not be interpreted, it is presumed turned off!
    #
    "enable": False,
    ##############################################################################
    # The `name` is the 'pretty print' (host)name that will be published.
    # spaces, apostrops are allowed, it is a (UTF-8) string!
    # if `service_name` is not provided here, the system's hostname (not so pretty)
    # will be pasted in.
    #
    # Notes:
    #  * if no service `name` is provided, the service will not start!
    #
    "name": socket.gethostname() or platform.node(),
    ##############################################################################
    # The `guest_account` keyword determines as what guest `user` the sksd
    # will spin up the Spyder Kernels.
    #
    # If there is no `guest_account`, this means that the user needs to provide
    # valid credentials for this host system.
    #
    # Notes:
    #  * if the guest account needs a password, the format is "user:password"
    #  * if the guest account needs an empty string password, the format is "user:"
    #  * if the guest account doesn't need a password, the format is "user"
    #
    "guest_account": "sct",
    ##############################################################################
    # The `guest_can_manage_environments` keyword determines if a `guest`
    # is allowed to create/modify (guest) environments.
    #
    # Notes:
    #  * if no guest_account is provided, this option is disregarded.
    #  * a `user` (guest or not) can in any case not create/modify environments
    #    outside the scope of the user !
    #
    "guest_can_manage_environments": False,
    ##############################################################################
    # The `max_kernels` keyword determines the maximum number of spyder-kernels
    # `skd` can spinn up. This is a (positive) integer value. If the value can't
    # be interpreted, the fallback is '0'. '0' means automatic calculation, it
    # takes the number of CPU cores, deducts 1 and that is set. There is in
    # principle no upper limit ...
    #
    # Notes:
    #  * `skd` will publish the value for `max_kernels` as well as the number of
    #    kernels already spinned up.
    #  * AUTO means # CPU cores - 1
    #  * anything starting with 'auto' (regardless capitalization) is considered 'AUTO'
    #  * CORES = the number of CPU cores
    #  * simple calculations are possible the result will be casted to integer.
    #    eg: (CORES/2)-1 ➜ (8/2)-1 ➜ 3 for a 8 core machine.
    #
    # Default value = AUTO
    #
    "max_kernels": AUTO,
    ##############################################################################
    # The `exclude_environments` keyword determines what environments will **NOT**
    # be published. (regardless if they exist or not)
    #
    # Notes:
    #  * It **must** be a list
    #  * One can use a wildcard '*' in the list.
    #  * application environments (starting with '_') can also be excluded here.
    #  * If nothing is provided, the default value is presumed.
    #
    # Default value = ["base", "_*"]
    #
    "exclude_environments": ["base", "_*"],
}


def config_path():
    """
    Return the path of the system configuration file.

    Returns
    -------
    str:
        Path to configuration file.
    """
    if sys.platform.startswith("linux"):
        path = "/etc"
    else:
        path = os.path.expanduser("~/.spyder-remote-server")

    if not os.path.isdir(path):
        os.makedirs(path)

    return os.path.join(path, INIFILE)


def create_config(guest, cores):
    """
    Create a confguration file in the configuration folder.

    Parameters
    ----------
    guest: str
        Account to use as guest account.
    cores: int
        Maximum number of spyder-kernels allowed.
    """
    config = DEFAULT_CONFIG.copy()
    config["guest_account"] = guest
    config["max_kernels"] = cores
    config["enable"] = True

    env = Environment(loader=PackageLoader("spyder_remote_server", "templates"))
    template = env.get_template(CONFIG_TEMPLATE)
    data = template.render(config=config)

    with open(config_path(), "w") as fh:
        fh.write(data)


def remove_config():
    """
    Remove configuration file from configuration path.
    """
    try:
        os.remove(config_path())
    except Exception as e:
        print(e)


def read_config():
    """
    Read Spyder Remote configuration.

    If no configuration is found, the default configuration is returned.

    Returns
    -------
    dict
        Configuration dictionary.
    """
    fpath = config_path()
    config_parser = configparser.ConfigParser()
    if os.path.isfile(fpath):
        config_parser.read(fpath)
        config = {}
        for key, default_value in DEFAULT_CONFIG.items():
            value = config_parser["DEFAULT"][key]
            if isinstance(default_value, (list, float, bool, int)):
                value = ast.literal_eval(value)

            config[key] = value
    else:
        config = DEFAULT_CONFIG.copy()

    return config


def get_log_path():
    """
    Get the logging folder path.

    Returns
    -------
    str
        Path to logging folder.
    """
    path = os.path.expanduser("~/.spyder-remote")
    if not os.path.isdir(path):
        os.makedirs(path)

    return os.path.join(path, "spyder-remote-server.log")



# Copyright (c) Semi-ATE
# Distributed under the terms of the MIT License
"""
Spyder Remote command line interface.
"""

import os
import subprocess
import sys

import click
from daemoniker import send, SIGINT
from spyder_remote_server.api import get_cpu_count, install_daemon, uninstall_daemon
from spyder_remote_server.constants import AUTO, PID_FILE_PATH
from spyder_remote_server.utils import is_root

# Constants
HERE = os.path.abspath(os.path.dirname(__file__))
DAEMON_FILE = "daemon.py"
DAEMON_PATH = os.path.join(HERE, DAEMON_FILE)


@click.group(
    help=(
        "Spyder Remote Server provides Spyder IDE kernels for conda "
        "environments using zeroconf."
    )
)
def main():
    pass


@main.command(help=("Install the server daemon"))
@click.argument("guest", type=click.STRING)
@click.argument("cores", type=click.INT, default=AUTO)
def install(guest, cores):
    """
    --guest [guest_account] it will also add a guest account to the system
    (with the same password as the account) and add this info also to the
    `/etc/spyder-remote.conf` file.

    Cores: and it will set the maximum numbers of consoles this machine will provide.
    """
    if not is_root():
        click.echo("Must install daemon as root", err=True)
        sys.exit(1)

    # click.echo("Install the server daemon")
    new_cores = get_cpu_count(cores)
    if cores != new_cores:
        click.echo(f"Using '{new_cores}' cores!")

    try:
        install_daemon(guest, new_cores)
    except PermissionError as e:
        click.echo("You need to install the daemon with sudo!")


@main.command(help=("Uninstall the server daemon"))
def uninstall():
    # click.echo("Uninstall the server daemon")

    try:
        uninstall_daemon()
    except PermissionError as e:
        click.echo("You need to uninstall the daemon with sudo!")
    except FileNotFoundError as e:
        pass


@main.command(help=("Start the server daemon"))
def start():
    # click.echo("Uninstall the server daemon")
    try:
        subprocess.Popen([sys.executable, DAEMON_PATH])
    except Exception as e:
        print(e)
        click.echo("Daemon already started!")


@main.command(help=("Stop the server daemon"))
def stop():
    try:
        # Send a SIGINT to a process denoted by a PID file
        send(PID_FILE_PATH, SIGINT)
        click.echo("Daemon stopped!")
    except Exception as e:
        click.echo("Daemon not started!")
